/*
 * Copyright 2009 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.asual.lesscss;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.URL;
import java.net.URLConnection;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jruby.Ruby;
import org.jruby.RubyRuntimeAdapter;
import org.jruby.RubyString;
import org.jruby.javasupport.JavaEmbedUtils;

/**
 * @author Rostislav Hristov
 */
public class Engine {

	private Ruby runtime;
	private RubyRuntimeAdapter adapter;
	private EngineCache cache;
	
	private final Log logger = LogFactory.getLog(getClass());
	
	public Engine() {
		
		List<String> loadPaths = new ArrayList<String>();
		loadPaths.add("META-INF/jruby.home/lib/ruby/site_ruby/1.8");
		loadPaths.add("META-INF/lesscss/lib");
		
		runtime = JavaEmbedUtils.initialize(loadPaths);
		adapter = JavaEmbedUtils.newRuntimeAdapter();
		cache = new EngineCache();
	}
	
	public void destroy() {
		JavaEmbedUtils.terminate(runtime);
	}
	
	public String parse(String source) throws EngineException {
		
		try {
			return ((RubyString) adapter.eval(runtime, 
					"require 'less' \nLess::Engine.new('" + source.replaceAll("'", "\"") + "').to_css")).toString().trim();
		} catch (Exception e) {
			logger.error("Compilation error.");
			throw new EngineException("Compilation error.", e);
		}
	}
	
	public String compile(File source) throws EngineException, IOException {
		return compile(source, cache);
	}
	
	public String compile(URL source) throws EngineException, IOException {
		return compile(source, cache);
	}
	
	private String compile(File source, EngineCache cache) throws EngineException, IOException {
		
		String path = source.getAbsolutePath();
		String folder = path.replace(source.getName(), "");
		
		if (source.exists()) {
			
			Long lastModified = source.lastModified();
			if (!cache.contains(path) || (cache.contains(path) && cache.getLastModified(path) < lastModified) || (areImportsModified(cache.getImports(path)))) {
				
				String content = readFile(source);
				
				Pattern p = Pattern.compile("@import\\s+(\"[^\"]*\"|'[^']*')");
				Matcher m = p.matcher(content);
				
				while (m.find()) {
					compile(new File(folder + m.group(1).replaceAll("\"|'", "")), cache.getImports(path));
				}
				
				if (cache == this.cache) {
					logger.info("Compiling '" + path + "'.");
					cache.setContent(path, ((RubyString) adapter.eval(runtime, 
							"require 'less' \nLess::Engine.new(File.new('" + path + "')).to_css")).toString().trim()).setLastModified(path, lastModified);
				} else {
					cache.setContent(path, content).setLastModified(path, lastModified);
				}
			}
			
			return cache.getContent(path);
			
		} else {
			logger.error("The file '" + path + "' doesn't exist.");
			throw new IOException("The file '" + path + "' doesn't exist.");
		}
	}
	
	private String compile(URL source, EngineCache cache) throws EngineException, IOException {
		
		if (source != null) {

			String path = source.getFile();			
			String folder = path.substring(0, path.lastIndexOf(System.getProperty("file.separator")) + 1);

			Long lastModified = source.openConnection().getLastModified();
			
			if (!cache.contains(path) || (cache.contains(path) && cache.getLastModified(path) < lastModified) || (areImportsModified(cache.getImports(path)))) {
				
				String content = readURL(source);
				
				Pattern p = Pattern.compile("@import\\s+(\"[^\"]*\"|'[^']*')");
				Matcher m = p.matcher(content);
				
				while (m.find()) {
					compile(new File(folder + m.group(1).replaceAll("\"|'", "")), cache.getImports(path));
				}
				
				if (cache == this.cache) {
					logger.info("Compiling: " + path);
					cache.setContent(path, ((RubyString) adapter.eval(runtime, 
							"require 'less' \nLess::Engine.new(File.new('" + path + "')).to_css")).toString().trim()).setLastModified(path, lastModified);
				} else {
					cache.setContent(path, content).setLastModified(path, lastModified);
				}
			}
			return cache.getContent(path);
			
		} else {
			logger.error("The requested URL doesn't exist.");
			throw new IOException("The requested URL doesn't exist.");
		}
	}
	
	private boolean areImportsModified(EngineCache cache) {
		for(String key : cache.keySet()) {
			if ((new File(key)).lastModified() > cache.getLastModified(key) || areImportsModified(cache.getImports(key))) {
				return true;
			}
		}
		return false;
	}
	
	private String readURL(URL source) throws IOException {
		
		StringBuilder sb = new StringBuilder();
		URLConnection urlc = source.openConnection();
        BufferedReader br = new BufferedReader(new InputStreamReader(urlc.getInputStream()));
        
		try {       
			String line = "";
	        while (line != null) {
	            sb.append(line);        
	            line = br.readLine();
	        }
		} catch (IOException e) {
			logger.error("Can't read '" + source.getFile() + "'.");
			throw e;
		}
		
		return sb.toString();
	}
	
	private String readFile(File source) throws IOException {

		String separator = System.getProperty("line.separator");
		StringBuilder sb = new StringBuilder();
		
		try {
			BufferedReader input = new BufferedReader(new FileReader(source));
			try {
				String line = null;
				while ((line = input.readLine()) != null) {
					sb.append(line).append(separator);
				}
			} finally {
				input.close();
			}
		} catch (IOException e) {
			logger.error("Can't read '" + source.getAbsolutePath() + "'.");
			throw e;
		}
		
		return sb.toString();
	}	
}