/*
 * Copyright 2009 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.asual.lesscss;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.util.Date;
import java.util.zip.GZIPOutputStream;

import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * @author Rostislav Hristov
 */
public class EngineServlet extends HttpServlet {

	private static final long serialVersionUID = 1L;
	private final Log logger = LogFactory.getLog(getClass());
	private Engine engine;
	private int cache = 31556926;
    private int milliseconds = 1000;

    @PostConstruct
    public void init() {
    	if (engine == null) {
			engine = new Engine();
    	}
	}

    @PreDestroy
    public void destroy() {
    	if (engine != null) {
    		engine.destroy();
    	}
	}
	
	public final void service(HttpServletRequest request,
			HttpServletResponse response) throws ServletException, IOException {
		
		URL url = getClass().getClassLoader().getResource("META-INF" + request.getRequestURI());
		if (url == null) {
			url = getClass().getClassLoader().getResource("META-INF/resources" + request.getRequestURI());
		}

		File file = null;
		try {
			file = new File(getServletContext().getRealPath(request.getRequestURI()));
		} catch (Exception e) {}
		
        if (url != null || (file != null && file.exists())) {
        	
        	Long lastModified = url != null ? url.openConnection().getLastModified() : file.lastModified();
            Long ifModifiedSince = request.getDateHeader("If-Modified-Since");
            
            if (ifModifiedSince != null && ifModifiedSince/milliseconds == lastModified/milliseconds) {
                response.setStatus(HttpServletResponse.SC_NOT_MODIFIED);
                return;
            }
            
            String compiled = null;
			try {
				compiled = url != null ? engine.compile(url) : engine.compile(file);
			} catch (EngineException ce) {
				logger.error("Error processing: " + request.getRequestURI());
				throw new ServletException(ce.getMessage(), ce);
			}

    		response.setContentType("text/css");
            response.addDateHeader("Expires", new Date().getTime() + cache*milliseconds);
            response.addDateHeader("Cache-control: max-age=", cache);
            response.addDateHeader("Last-Modified", lastModified);
            
            if ("true".equals(getInitParameter("gzip"))) {
            	response.setHeader("Content-Encoding", "gzip");
            	GZIPOutputStream gz = new GZIPOutputStream(response.getOutputStream());
                gz.write(compiled.getBytes());
            	gz.close();
            } else {
				response.getOutputStream().print(compiled);
            }

        	response.getOutputStream().close();
            
        } else {
        	logger.error("Error processing: " + request.getRequestURI());
			response.setStatus(HttpServletResponse.SC_NOT_FOUND);
        }
	}	
}